// -----------------------------------------------------------
// Code written by Cristiano Tapparello
// Department of Information Engineering, University of Padova
//
// Contact:	tappare1@dei.unipd.it
// Original Version: August 2008
// Updated on July 2012
// -----------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <boost/dynamic_bitset.hpp>

#include <vector>
#include <iostream>

#include "exp_statistics_opt.c"

#include "hashlib.c"

#define sgn(x) (x)==0.0?0.0:x/fabs(x)
#define MAX(x,y) (x>=y ? x : y)
#define MIN(x,y) (x>=y ? y : x)
#define MIN3(x,y,z) MIN(MIN(x,y),z)

#define DEBUG_FRTDP 0
#define DEBUG_UL 0
#define DEBUG_PRUNE 0
#define DELAY 1

// set gamma inside (0,1)
#define gamma 0.99

#define MAX_DELAY 1000
#define MAX_RETRANS 10000

#define H_LOWER 0
//#define H_UPPER 100

#define Pr	10
#define kappa	3.5

long double Y = 0.000015088;  //0.000015088  80% @ 30 meters // 0.000705641  80% a 10 m   // 0.000000223 80% a 100m
unsigned NUMBER_OF_NODES;

double H_UPPER;

// Maximum number of transmitting node per round
unsigned MAX_TRANS;

// Etx/Erx set omega [0,1]
double omega;

// cost function weighting parameter
double ALPHA;

// set DELTA >= 0
double DELTA = 1;

// set epsilon >= 0 (quality of the solution) 
double epsilon = 0.01;

// choose best action according to Lower bound (1) or Upper bound (0)
unsigned LOWER = 1;

double MAX_X;

using namespace std;

struct STATE {
    boost::dynamic_bitset<> set;
    boost::dynamic_bitset<> astar;
    boost::dynamic_bitset<> sstar;
    
    vector< boost::dynamic_bitset<> > next;
    vector<boost::dynamic_bitset<> > actions;
    
	double	upper;
	double	lower;
	double	delta;
	double  priority;
	long    n_next_states;
	long    n_actions;
	double sum_probability;
	double visit;
};

typedef std::pair<int, long double> NODE;

//********************************************************************************//
//  Called everytime we need to create a new state                                //
//********************************************************************************//
STATE *state_init(int N){
    STATE *ret = new STATE;
    ret->set = boost::dynamic_bitset<>(N);
    ret->astar = boost::dynamic_bitset<>(N);
    ret->sstar = boost::dynamic_bitset<>(N);
    
    ret->lower = H_LOWER;
    ret->upper = H_UPPER;
    ret->delta = 0;
    ret->n_next_states = 0;
    ret->n_actions = 0;
	ret->visit = 0;
    ret->priority = fabs(H_UPPER - H_LOWER) - epsilon/2;
    return ret;
}

//********************************************************************************//
//  Called everytime we need to release the memory space of a state               //
//********************************************************************************//
void state_free(STATE * release){

    ~(release->set);
    ~(release->astar);
    ~(release->sstar);
    
    for (int i=0;i<release->n_next_states;i++){
        ~(release->next[i]);
    }
    release->next.clear();
    
    for (int i=0;i<release->n_actions;i++){
        ~(release->actions[i]);
    }
    release->actions.clear();
    delete release;
}

//*******************************************************************************//
// Hashlib Auxiliary functions BEGIN                                             //
//*******************************************************************************//

// function to compare two elements of the hash table 0 = equal, 1 = different
int mycmp(void *litem, void *ritem){
    STATE * left = (STATE *)litem;
    STATE * right = (STATE *)ritem;
    return  left->set != right->set;
}

// function to store an element in the hash table
void *mydupe(void *item){
    STATE *added = (STATE *)item;
    return added;
}

// function to release a new element in the hash table
void myundupe(void *item){
    state_free((STATE *)item);
}

unsigned long myhash(void *item){
    STATE * myitem = (STATE *)item;
    return myitem->set.to_ulong();
}

unsigned long myrehash(void *item){
    STATE * myitem = (STATE *)item;
    return myitem->set.to_ulong()>>3;
}
//*******************************************************************************//
// Hashlib Auxiliary functions  END                                              //
//*******************************************************************************//


//*******************************************************************************//
// Compare two NODEs according to some criterion (e.g., distance to destination, //
// success probability, ...)                                                     //
// return   1 if (i < j)                                                         //
//          0 otherwise                                                          //
//*******************************************************************************//
bool compare_nodes (NODE i, NODE j){
	return (i.second < j.second);
}

bool isFinal(boost::dynamic_bitset<> set){
    return set[0];
}

double sigma(double distance){
    double sigma;
    sigma = (Pr * pow(distance,-kappa));
    return sigma;
}

// Global variables used frequently in the program so we
// create only once to save time
boost::dynamic_bitset<> RECEIVE,SSTAR,A_MAX,V,N,NEXT_SET,NEXT_NODE;
STATE *to_search;

vector<vector<long double> > H;
vector<vector<double> > d;

long double *alpha;
double delta, qcurr, qprev,kD, D;
int ncurr,nprev;
unsigned long long state_analized;
hshtbl * state_collection;
STATE **state_visited;

double tot_visited;
double ALPHA_TO_1;
double cost_value;
long int delay_value;


//********************************************************************************//
//  Calculate the probability of going from state CURRENT to state NEXT using     //
//  using action ACTION                                                           //
//********************************************************************************//
double trans_prob(boost::dynamic_bitset<> current, boost::dynamic_bitset<> next, boost::dynamic_bitset<> action){
    long double probability = 1;
    unsigned i,j,n_sender,block;
    boost::dynamic_bitset<> RECEIVE = current ^ next;
	
	long double tmp;
	
    //compute probability
    for (i=0;i<NUMBER_OF_NODES;i++){
        if (!current[i]){
            n_sender = 0;
            if (RECEIVE[i]){
			    block = 0;
                for (j=0;j<NUMBER_OF_NODES;j++){
                    if (action[j]){
						if (H[i][j] < ALPHA_TO_1){
							block = 1;
						}
						alpha[n_sender++]=H[j][i];
                    }
                }
				if ((block) && (n_sender>3)) tmp = 0.999999;
				else tmp = exp_ccdf(n_sender,alpha,Y);
				
				if ((tmp<0) || (tmp>1)) {
                    cout << "current " << current;
                    cout << " next " << next;
                    cout << " receive " << RECEIVE;
                    cout << " action " << action;
					for (j=0;j<n_sender;j++){
                        cout << " alpha " << j << " " << alpha[j] << endl;
					}
                    cout << " action transmit to " << i << " " << tmp << endl;
					exit(100);
				}
				probability *= tmp;
            } else {
				block = 0;
                for (j=0;j<NUMBER_OF_NODES;j++){
                    if (action[j]){
						if (H[i][j] < ALPHA_TO_1){
							block = 1;
						}
                        alpha[n_sender++]=H[j][i];
					}
                }
				if ((block) && (n_sender>3)) tmp = 0.000001;
				else tmp = 1 - exp_ccdf(n_sender,alpha,Y);
				if ((tmp<0) || (tmp>1)) {
                    cout << "current " << current;
                    cout << " next " << next;
                    cout << " receive " << RECEIVE;
                    cout << " action " << action;
					for (j=0;j<n_sender;j++){
                        cout << " alpha " << j << " " << alpha[j] << endl;
					}
                    cout << " action do not transmit to " << i << " " << 1.0-tmp << endl;
					exit(100);
				}
				
                probability *= tmp;
            }
        }
    }
    
    return probability;
}

//********************************************************************************//
//  Calculate the cost of going from state current to state next using action     //
//  action                                                                        //
//********************************************************************************//
double cost(boost::dynamic_bitset<> current, boost::dynamic_bitset<> next, boost::dynamic_bitset<> action){
    double cost = 0.0;
	cost = (double) action.count() + omega *(next.count() - current.count());
    cost = (1.0 - ALPHA) + ALPHA*cost;
    return cost;
}

//********************************************************************************//
//  Determine and prune the set of next states                                    //
//********************************************************************************//
void prune(STATE *s){
    if (!s->next.empty()) return;
    
    vector<NODE> actions;
    vector<NODE> v;
    
    int i,k,j,n_sender,max_receiver;
    max_receiver = NUMBER_OF_NODES - s->set.count();
    
    for (i=1;i<NUMBER_OF_NODES;i++){
        if (s->set[i]){
            actions.push_back(NODE(i,(long double)d[i][0]));
        }
    }
    
    sort(actions.begin(), actions.end(),compare_nodes);
    
    n_sender = MIN(s->set.count(),MAX_TRANS);
    unsigned long n_actions = (unsigned long) pow(2,n_sender)-1;
    s->n_actions = n_actions;
    
    A_MAX.reset();
    
    for (i=0;i<n_sender;i++){
        A_MAX[actions[i].first] = 1;
    }
    
	if (DEBUG_PRUNE) {
        cout << "max transmitter " << A_MAX << endl;
	}
	
    // for every node that still have to decode the message, calculate the probability
    // that this node will correctly decode the message if nodes in set 'A_MAX' will
    // transmit
    for (k=0;k<NUMBER_OF_NODES-1;k++){
        if (!s->set[k]){
            n_sender=0;
            for (j=1;j<NUMBER_OF_NODES;j++){
                if (A_MAX[j]){
                    alpha[n_sender++]=H[j][k];
                }
            }
            v.push_back(NODE(k,exp_ccdf(n_sender,alpha,Y)));
            
        }
    }
    
    sort(v.begin(), v.end(),compare_nodes);
    
    V.reset();
    double M_ = DELTA / ((1+omega)*(NUMBER_OF_NODES-1)+gamma*H_UPPER);
    
    double Mx = 0;
    double productory=1;
    k = 0;
    
    do {
        Mx = Mx*(1-v[k].second) + v[k].second * productory;
        if ((Mx<=M_) && (k<max_receiver-1)){
            V[v[k].first] = 1;
            productory *= (1-v[k].second);
            k++;
        }
    } while ((Mx<=M_) && (k<max_receiver-1));
    
    
    // GENERATE ALL POSSIBLE PRUNED NEXT STATES
    unsigned long n_next_states = (unsigned long) pow(2,max_receiver-k);
    s->n_next_states = n_next_states;
    for (int i=0;i<n_next_states;i++) s->next.push_back(boost::dynamic_bitset<>(NUMBER_OF_NODES));
	
    N = ~(s->set | V);
    
	if (DEBUG_PRUNE){
        cout << "node not to transfer to " << V << endl;
        cout << "node to tranfer to " << N << " " << n_next_states << endl;
	}
    
    int h, n_1, n_0;
    h = n_1 = n_0 = 1;
    for (i = 0; i< NUMBER_OF_NODES ; i++){
        if (s->set[i]) {
            for (j=0; j<n_next_states;j++) {
                s->next[j][i] = 1;
            }
        }
        else
            if (N[i]){
                for (j=0; j<n_next_states;j++) {
                    if (n_1 > 0) {
                        n_1--;
                        s->next[j][i] = 1;
                    }
                    else {
                        n_0--;
                        if (n_0 == 0) {
                            n_1 = n_0 = h;
                        }
                    }
                }
                h *= 2;
                n_1 = n_0 = h;
            }
    }
    
    // GENERATE ALL POSSIBLE PRUNED ACTIONS
    for (i=0;i<n_actions;i++) s->actions.push_back(boost::dynamic_bitset<>(NUMBER_OF_NODES));
    
    h = n_1 = n_0 = 1;
    for (i = 1; i< NUMBER_OF_NODES ; i++){
        if (A_MAX[i]){
            for (j=0; j<n_actions;j++) {
                if (n_1 > 0) {
                    n_1--;
                    s->actions[j][i] = 1;
                }
                else {
                    n_0--;
                    if (n_0 == 0) {
                        n_1 = n_0 = h;
                    }
                }
            }
            h *= 2;
            n_1 = n_0 = h;
        }
    }
	
	if (DEBUG_PRUNE){
        cout << "state " << s->set <<endl;
		for (i=0;i<n_actions;i++){
            cout << "action " << s->actions[i] << endl;
		}
		for (i=0;i<n_next_states;i++){
            cout << "next state " << s->next[i] << endl;
		}
	}
    
}

//********************************************************************************//
//  Calculate the upper bound, the lower bound and the relative best action       //
//  of a particular state s                                                        //
//********************************************************************************//
void V_upper_and_lower(STATE *s){
    double next_upper, next_lower, next_priority,p,sum_probability,c_cost;
    double upper, lower, probability, priority, min_upper,min_lower,max_priority;
    int action,next;
    min_lower = H_UPPER * 10;
    min_upper = H_UPPER * 10;
    STATE * next_s;
	state_analized += s->n_next_states;
    for (action=0; action< s->n_actions; action++){
        upper = lower = sum_probability = 0;
        max_priority = - epsilon/2;
        for (next=0; next<s->n_next_states;next++){
            if (!isFinal(s->next[next])){
                to_search->set = s->next[next];
                next_s = (STATE *)hshfind(state_collection,to_search);
                if (next_s != NULL) {
                    next_upper = next_s->upper;
                    next_lower = next_s->lower;
                    next_priority = next_s->priority;
                } else {
					next_upper = H_UPPER;
					next_lower = H_LOWER;
					next_priority = fabs(H_UPPER - H_LOWER) - epsilon/2;
				}
            }
            else {
                next_upper = 0;
                next_lower = 0;
                next_priority = - epsilon/2;
            }
            probability = trans_prob(s->set,s->next[next],s->actions[action]);
            c_cost = cost(s->set,s->next[next],s->actions[action]);
            sum_probability += probability;
            upper += probability * (c_cost + gamma * next_upper);
            lower += probability * (c_cost + gamma * next_lower);
			
			if (DEBUG_UL){
                cout <<"state " << s->set;
                cout << " action " << s->actions[action]; 
                cout << " next " << s->next[next] << " " << probability << endl; 
			}
            
            priority = gamma * probability * next_priority;
            
            // search for next state that maximize priority
            if (priority > max_priority) {
                SSTAR = s->next[next];
                max_priority = priority;
            }
        } // end -> for (next=0; next<s->n_next_states;next++)
        
		upper = upper/sum_probability;
        lower = lower/sum_probability;
		
        max_priority = max_priority/sum_probability;
        
		if (DEBUG_UL){
			cout <<"state " << s->set;
            cout << " action " << s->actions[action] << " " << lower << endl;
		}
		
		// check if current action gives the best lower bound and in case store the value and
        // the bitset relative to the action
        if (lower < min_lower) {
            min_lower = lower;
            if (LOWER == 1){
                s->astar = s->actions[action];
                s->sstar = SSTAR;
                p = max_priority;
                s->sum_probability = sum_probability;
			}
        }
        
        if (upper < min_upper){
            min_upper = upper;
			if (LOWER == 0){
				s->astar = s->actions[action];
				s->sstar = SSTAR;
				p = max_priority;
				s->sum_probability = sum_probability;
			}
        }
    }
    
    //update lower bound
    s->lower = min_lower; 
    delta = fabs(s->upper - min_upper);
    //update upper bound
    s->upper = min_upper; 
    s->delta = fabs(min_upper - min_lower) - epsilon/2;
    s->priority = MIN(s->delta,p);
}

//********************************************************************************//
//  Update the upper and lower bounds of a state and chose the next state         //
//  to process                                                                    //
//********************************************************************************//
void backup(STATE *s){
    prune(s);
    //compute the new upper bound,lower bound and the relative best action relative to state s
    V_upper_and_lower(s);
}


//********************************************************************************//
//  Regulate the track update quality                                             //
//********************************************************************************//
void trackUpdateQuality(double q, int d){
    if (d>D/kD) {
        qcurr += q;
        ncurr++;
    } else{
        qprev += q;
        nprev++;
    }
}


//********************************************************************************//
//  Recursive step of the algorithm                                               //
//********************************************************************************//
void trialRecurse(STATE *s, double W, int d){
    if (isFinal(s->set)) return;
    backup(s);
    trackUpdateQuality(delta*W,d);
    
	if (DEBUG_FRTDP){
        cout <<"state " << s->set;
        cout << " action " << s->astar; 
        cout << " next " << s->sstar << endl;
	}
    
    if ((s->delta <=0) || (d>=D)) return;
    
    STATE * next = NULL;
    
    to_search->set = s->sstar;
    
    // check if s->sstar has already been processed
    next = (STATE *) hshfind(state_collection,to_search);
    // otherwise create a new state and add it to the collection
    if (next == NULL) {
        state_analized++;
		if (DEBUG_FRTDP){
            cout << "state not found, create it and add to the collection " << s->sstar << endl;
        }
		next = state_init(NUMBER_OF_NODES);
        next->set = s->sstar;
        if (next->set[0]){
            next->upper = next->lower = 0;
            next->delta = -epsilon/2;
            next->priority = -epsilon/2;
        }
        hshinsert(state_collection,next);
    }
    
    trialRecurse(next,gamma*trans_prob(s->set,s->sstar,s->astar)/s->sum_probability*W,d+1);
    backup(s);
}

// Count cardinality of best action
int execfn(void *item, void *datum, void *extra){
    double * xdata = (double *)datum;
    STATE *xitem= (STATE *)item;
    xdata[xitem->astar.count()]+= (double)((xitem->visit / tot_visited));
    return 0;
}


//********************************************************************************//
//  Calculate the delay of the message by simulating the optimal policy.          //
//********************************************************************************//
void calculate_delay(){
	double p_rec,prob;
	int n_sender,j;
	unsigned received;
	int not_received = 0;
    
    to_search->set.reset();
    to_search->set[NUMBER_OF_NODES-1] = 1;
	STATE * actual = NULL;
	actual = (STATE *)hshfind(state_collection, to_search);
	int pos;
	delay_value = 0;
	cost_value = 0;
	int state_seen = 0;
	
	while (!actual->set[0]){
        NEXT_SET.reset();
		received = 0;
		for (pos =0;pos<NUMBER_OF_NODES;pos++){
            NEXT_NODE.reset();
			if(actual->set[pos]){
                NEXT_SET[pos] = 1;
			} else {
                NEXT_NODE[pos] = 1;
				p_rec  = (double)rand()/RAND_MAX;
				
				// take alpha between node 'pos' and transmitting nodes
				n_sender=0;
				for (j=1;j<NUMBER_OF_NODES;j++){
					if (actual->astar[j]){
						alpha[n_sender++]=H[j][pos];
					}
				}
				
				prob = exp_ccdf(n_sender,alpha,Y);
				
                if (p_rec <= prob){
					received = 1;
                    NEXT_SET[pos] = 1;
				}
			}
		}
		delay_value++;
		cost_value += (double)n_sender + omega *(NEXT_SET.count() - actual->set.count());
		// check if next_set is a final state
		if (NEXT_SET[0]){
			break;
		}
		// check if someone has received the message
		if (received){
			to_search->set = NEXT_SET;
			// find state with set 'next_set' 
			actual = (STATE *) hshfind(state_collection, to_search);
			if (actual == NULL) {
				// next state not foud, discard computation
				delay_value = 0;
				cost_value = 0;
				return;
			}
			// next state found, add to state_visited and increment delay 
			state_visited[state_seen++] = actual;
		} else {
			// if not received more than 10000 times, discard computation
			if (++not_received>MAX_RETRANS){
				delay_value = 0;
				cost_value = 0;
				return;
			}
		}
	}
	
	// Delay successfully determined, increment the number of times we have seen each state 
	// and the total number of state visited so far
	for (j=0;j<state_seen;j++){
		state_visited[j]->visit +=1;
		tot_visited += 1;
	}
    
}


//********************************************************************************//
//  Algorithm Main                                                                //
//********************************************************************************//
int main(int argc, char **argv){
    
	// argv[1] = Number of nodes in the network
	// argv[2] = Epsilon
	// argv[3] = DELTA
	// argv[4] = MAX_TRANS
	// argv[5] = MAX_X
	// argv[6] = MAX_Y
	// argv[7] = ALPHA
	// argv[9] = H_HUPPER
	// argv[10] = filename
	// argv[11] = n_sample
	
    srand((unsigned)time(NULL));
    unsigned i,j;
	int n_sample,iterate_random;
	double mean_activation;
	
	ALPHA_TO_1 = 53;
	
    kD = 1.1;
    FILE *stream;
	FILE *streamActivation;          
	char str1[100] = "Action_";    
    stream = fopen(argv[9], "a");
	strcat(str1,argv[9]);
	printf("%s\n",str1);
	streamActivation = fopen(str1, "a");
    NUMBER_OF_NODES = atoi(argv[1]);
	epsilon = atof(argv[2]);
	DELTA = atof(argv[3]);
    MAX_TRANS = atoi(argv[4]);
    double MAX_X = atof(argv[5]);
	double MAX_Y = atof(argv[6]);
	ALPHA = atof(argv[7]);
    H_UPPER = atof(argv[8]);
	n_sample = atoi(argv[10]);
    
    //other parameters
    omega = 0;
    long rep_delay = 1000001;
    long MAX_FAILURE = 20000;
    
	// Create global variables
    V = boost::dynamic_bitset<>(NUMBER_OF_NODES);
    N = boost::dynamic_bitset<>(NUMBER_OF_NODES);
    A_MAX = boost::dynamic_bitset<>(NUMBER_OF_NODES);
	SSTAR = boost::dynamic_bitset<>(NUMBER_OF_NODES);
    RECEIVE = boost::dynamic_bitset<>(NUMBER_OF_NODES);
	to_search = state_init(NUMBER_OF_NODES);
	NEXT_NODE = boost::dynamic_bitset<>(NUMBER_OF_NODES);
	NEXT_SET = boost::dynamic_bitset<>(NUMBER_OF_NODES);
    H = vector<vector<long double> >(NUMBER_OF_NODES, vector<long double>(NUMBER_OF_NODES,0.0L)); 
    d = vector<vector<double> >(NUMBER_OF_NODES, vector<double>(NUMBER_OF_NODES,0.0));
    alpha = (long double *)malloc(sizeof(long double)*NUMBER_OF_NODES);
    state_visited = (STATE **)malloc(sizeof(STATE *)*MAX_DELAY);
	
    
	double total_mean_delay,total_mean_cost, mean_upper, mean_lower,mean_delay,mean_cost;
	int delay_sample,success,failure;
	
    vector<double> x(NUMBER_OF_NODES,0.0);
    vector<double> y(NUMBER_OF_NODES,0.0);
	
	int rangeX = (int)MAX_X + 1;
	int rangeY = (int)MAX_Y + 1;
	
	double * activation = (double *)malloc(sizeof(double)*(MAX_TRANS+1));
	
    iterate_random = 0;
    
	do{ // Iterate over random topologies
        // Destination position
        x[0] = MAX_X;
        y[0] = floor(MAX_Y / 2);
        // Source position
        x[NUMBER_OF_NODES-1] = 0;
        y[NUMBER_OF_NODES-1] = floor(MAX_Y / 2);
		// Position of other nodes in the scenario
		for (i = 1; i< NUMBER_OF_NODES-1;i++){
			x[i] = rand() % rangeX;
			y[i] = rand() % rangeY;
		} 
		
		for (i=0;i<NUMBER_OF_NODES;i++){
			fprintf(streamActivation, "(%f,%f)\n",x[i],y[i]);
		}
		
		for (i=0;i<NUMBER_OF_NODES;i++){
			d[i][i] = 0.0;
			H[i][i] = 0.0;
			for (j=i+1;j<NUMBER_OF_NODES ;j++ ){
				d[i][j] = d[j][i] = (long double)sqrt(pow((long double)(x[i]-x[j]),2) + pow((long double)(y[i]-y[j]),2));
				H[i][j] = H[j][i] = (long double)(1 / sigma(d[i][j]));
			}
		}
		// END create matrix
        
        fprintf(streamActivation,"nodes=  %d\t Max trans= %d\t omega= %f\t ALPHA= %f\t epsilon= %f\t DELTA= %f\t MAXx= %f\tMAXy= %f\t H_Up = %f\n",NUMBER_OF_NODES,MAX_TRANS,omega, ALPHA, epsilon, DELTA,MAX_X,MAX_Y,H_UPPER);
        printf("nodes=  %d\t Max trans= %d\t omega= %f\t ALPHA= %f\t epsilon= %f\t DELTA= %f\t MAXx= %f\tMAXy= %f\t H_Up = %f\n",NUMBER_OF_NODES,MAX_TRANS,omega, ALPHA, epsilon, DELTA,MAX_X,MAX_Y,H_UPPER);
        total_mean_delay = 0;
        mean_upper = 0;
        mean_lower = 0;
        
        
        state_collection = hshinit(myhash,myrehash,mycmp,mydupe,myundupe,0);
        
        // create state s0
        STATE *s0 = state_init(NUMBER_OF_NODES);
        s0->set[NUMBER_OF_NODES-1] = 1;
        state_analized = 1;
        hshinsert(state_collection,s0);
        
        int num_of_execution = 0;
        // HERE THE COMPUTATION REALLY STARTS
        D = 10;
        while ((s0->upper - s0->lower)>epsilon){
            qprev = 0;
            qcurr = 0;
            ncurr = 0;
            nprev = 0;
            printf ("number of executions %d upper %f, lower %f %f %llu\n",num_of_execution++,s0->upper,s0->lower,s0->upper - s0->lower,state_analized);
            trialRecurse(s0,1,0);
            if ((qcurr/ncurr)>=(qprev/nprev)) D = kD * D;
            if ((s0->upper - s0->lower)<epsilon) break;
        }
        printf ("upper %f, lower %f\t %llu\n",s0->upper,s0->lower,state_analized);
        
        tot_visited = 0;
        mean_delay = 0;
        mean_cost = 0;
        success = 0;
        failure = 0;
        
        if (DELAY){
            for (i=1;i<rep_delay;i++){
                calculate_delay();
                if (delay_value>0){
                    success++;
                    mean_cost = ((success-1)*mean_cost + cost_value)/success;
                    mean_delay = ((success-1)*mean_delay + delay_value)/success;
                }else{ 
                    i--;
                    if (++failure > MAX_FAILURE) i++;
                }
            }
            fprintf(stream,"%f\t%f\t%f\t%f\t%llu\t%d",ALPHA,mean_delay,mean_cost,s0->lower,state_analized,num_of_execution);
        }
        
        s0->visit = success;
        tot_visited += success;
        
        if (failure<MAX_FAILURE) {
            delay_sample++;
            total_mean_delay = (total_mean_delay*(delay_sample-1) + mean_delay)/delay_sample;
            total_mean_cost = (total_mean_cost*(delay_sample-1) + mean_cost)/delay_sample;
        }
        fprintf(streamActivation,"Failure %d success %d\n",failure,success);
		
        for (j=0;j<=MAX_TRANS;j++){
            activation[j] = 0.0;
        }
        
        hshwalk(state_collection, execfn, activation);
        
        mean_upper = s0->upper;
        mean_lower = s0->lower;
        
        mean_activation = 0;
        for (i=0;i<=MAX_TRANS;i++){
            fprintf(streamActivation,"%d activations \t%f\n",i,activation[i]);
            mean_activation += activation[i]*i;
        }
        
        fprintf(stream, "\t%f\n",mean_activation);
        
        hshkill(state_collection);
        state_collection = NULL;
        fflush(stream);
        fflush(streamActivation);
        
	} while (iterate_random++<n_sample);
	
    fflush(stream);
    fclose(stream);
	fflush(streamActivation);
    fclose(streamActivation);
    
    return 0;
}
